/***********************************************************************************************************************
 *
 * BlueBook - use live code examples in your book
 * Copyright (C) 2011-2011 by Tidalwave s.a.s.
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebook.java.net
 * SCM: https://hg.java.net/hg/bluebook~src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebook.plugin.maven;

import javax.annotation.Nonnull;
import lombok.AllArgsConstructor;
import lombok.Cleanup;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import it.tidalwave.bluebook.configuration.ConfigurationVisitor;
import it.tidalwave.bluebook.filter.java.JavaFilter;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import org.apache.commons.io.IOUtils;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.artifact.InvalidDependencyVersionException;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: MavenConfigurationVisitor.java,v 11dfcc8e5b11 2011/03/10 00:01:56 fabrizio $
 *
 **********************************************************************************************************************/
@AllArgsConstructor
public class MavenConfigurationVisitor extends ConfigurationVisitor
  {
    private Log log;
    
    private ArtifactFactory factory;

    private ArtifactResolver artifactResolver;

    private List remoteRepositories;

    private ArtifactRepository localRepository;
    
    private MavenProject mavenProject;
    
    private ArtifactMetadataSource artifactMetadataSource;
    
    private ProgramListingTemplate template = new ProgramListingTemplate();
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    @Override @Nonnull
    public ConfigurationVisitor clone()
      {
        final MavenConfigurationVisitor clone = (MavenConfigurationVisitor)super.clone();

        clone.log                    = this.log; 
        clone.factory                = this.factory;
        clone.artifactResolver       = this.artifactResolver;
        clone.remoteRepositories     = this.remoteRepositories;
        clone.localRepository        = this.localRepository;
        clone.mavenProject           = this.mavenProject;
        clone.artifactMetadataSource = this.artifactMetadataSource;

        return clone;
      }
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    @Override
    public void extractFile() 
      throws MojoExecutionException 
      {
        try 
          {
            log.info("Creating generated example snippets...");
            log.debug("configuration: " + this);
           
            if ("".equals(version))
              {
                searchVersionInDependencyManagement();
              }
            
            if ("".equals(version))
              {
                throw new MojoExecutionException("Can't find version of " + groupId + ":" + artifactId + " in dependencyManagement");  
              }
            
            final org.apache.maven.artifact.Artifact artifact = factory.createArtifactWithClassifier(groupId, artifactId, version, type, classifier);
            artifactResolver.resolve(artifact, remoteRepositories, localRepository);
                        
            final java.io.File artifactFile = artifact.getFile();
            log.info("artifact file: " + artifactFile);
            log.info("jar entry:     " + fileName);
            final JarFile jarFile = new JarFile(artifactFile);
            final JarEntry jarEntry = jarFile.getJarEntry(fileName);
            
            if (jarEntry == null)
              {
                throw new MojoExecutionException("Cannot find " + fileName + " in " + artifact);  
                // TODO: see http://java.net/jira/browse/BLUEBOOK-6
              }
            
            final boolean binary = isBinary(fileName);
            
            if ("".equals(outputFile))
              {
                outputFile = fileName.replaceAll("^.*/", "");
                
                if (!binary)
                  {
                    outputFile += ".xml";  
                  }
              }
            
            final StringBuilder buffer = new StringBuilder("target/generated-sources/docbkx/"); // FIXME
            
            if (!"".equals(outputDirectory))
              {
                buffer.append(outputDirectory).append(java.io.File.separator);  
              }
            
            final java.io.File targetFile = new java.io.File(buffer.append(outputFile).toString());
            log.info("Creating " + targetFile.getAbsolutePath() + "...");
            targetFile.getParentFile().mkdirs();
            
            if (binary)
              {
                processBinary(jarFile, jarEntry, targetFile);
              }
            else
              {
                processText(jarFile, jarEntry, targetFile);
              }
          }
        catch (InvalidDependencyVersionException e) 
          {
            throw new MojoExecutionException("", e);
          } 
//        catch (ProjectBuildingException e) 
//          {
//            throw new MojoExecutionException("", e);
//          }
        catch (IOException e) 
          {
            throw new MojoExecutionException("", e);
          }
        catch (ArtifactResolutionException e) 
          {
            throw new MojoExecutionException("", e);
          }
        catch (ArtifactNotFoundException e) 
          {
            throw new MojoExecutionException("", e);
          }
      }
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private void searchVersionInDependencyManagement() 
      throws InvalidDependencyVersionException, ArtifactResolutionException, ArtifactNotFoundException
      {
        final Set<?> dependencyArtifacts = mavenProject.createArtifacts(factory, null, null);
        final ArtifactResolutionResult result = artifactResolver.resolveTransitively(dependencyArtifacts, 
                                                                                     mavenProject.getArtifact(), 
                                                                                     remoteRepositories, 
                                                                                     localRepository, 
                                                                                     artifactMetadataSource);
        for (final Artifact depManArtifact : (Set<Artifact>)result.getArtifacts()) 
          {
            // do not use classifier and type in the comparison: so we can take advantage of the regular
            // default dependency (i.e. the jar) insteading of requiring one with specific classifier.
            //
            if (depManArtifact.getGroupId().equals(groupId) 
                && depManArtifact.getArtifactId().equals(artifactId))
              {
                version = depManArtifact.getVersion();
                log.info("Artifact version from dependencyManagement: " + version);
                break;
              }
          }
      }

    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private void processText (final @Nonnull JarFile jarFile, final @Nonnull JarEntry jarEntry, @Nonnull final java.io.File targetFile) 
      throws IOException 
      {        
        final @Cleanup Reader r = new InputStreamReader(jarFile.getInputStream(jarEntry));
        final @Cleanup FileWriter w = new FileWriter(targetFile);
        
        final JavaFilter javaFilter = new JavaFilter();
        javaFilter.setSkipBlockComments(skipBlockComments);
        javaFilter.setSkipHeader(skipHeader);
        javaFilter.setSkipImports(skipImports);
        javaFilter.setSkipPackage(skipPackage);
        javaFilter.setIncludedFragment(includedFragment);
        javaFilter.setExcludedFragment(excludedFragment);
        
        template.process(javaFilter, r, w);
      }
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private void processBinary (final @Nonnull JarFile jarFile, final @Nonnull JarEntry jarEntry, @Nonnull final java.io.File targetFile) 
      throws IOException 
      {        
        final @Cleanup InputStream is = jarFile.getInputStream(jarEntry);
        final @Cleanup OutputStream os = new FileOutputStream(targetFile);
        IOUtils.copy(is, os);
      }
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private boolean isBinary (final @Nonnull String fileName)
      {
        return !fileName.endsWith(".java"); // FIXME: of course, we need something better  
      }
  }
